# General function to perform experiments with RANKS
# July 2012

library(PerfMeas);
library(NetPreProc);
source("sw/ranks.R");

# do.RANKS
# High level function to perform experiments with RANKS
# It perform a 5 fold CV repeated 1 time and 10 times on a given data set
# INPUT:
# score : function. It must be a kernel-based score method:
#                 - KNN.score (default)
#                 - NN.score 
#                 - eav.score
# kernel : kernel method (def. rw.kernel)
# a : kernel parmater (def. 2)
# k : number of neighbours for KNN.score (def.19)
# p : number od steps of the RW kernel (def. 1)
# data : name of the data set to loaded (without rda extension). It must be  an .rda file containing the adjiacency matrix of the graph. 
#        It assumes that it is in the "data" directory
# labels : name of the target labels (without rda extension). It must be  an .rda file containing the label matrix of the examples.
#          It assumes that it is in the "data" directory
# OUTPUT:
# 3 rda files stored in the Results directory:
# - Scores results: A matrix with examples on rows and classes on columns representing the computed scores for each example and for each considered class
# - AUC results files computed through AUC.single.over.classes
# - Precision at given recall results computed through precision.at.multiple.recall.level.over.classes
do.RANKS  <- function(score=KNN.score, kernel=rw.kernel, a=2, k=19, p=1, data, labels)  {

  recall.levels <- c(0.01, 0.05, seq(from=0.1, to=1, by=0.1));
  
  # loading the adjacency matrix
  dataset.name <- paste0("data/",data,".rda");
  data.name=load(dataset.name);
  K=eval(parse(text=data.name));  # K represents the adjacency matrix
  K[K<0]<-0;

  # loading labels matrix
  dataset.name <- paste0("data/",labels,".rda");
  x=load(dataset.name);
  T=eval(parse(text=x));  # T represents the label matrix

  nclasses <- ncol(T);
  ndrugs <- nrow(T);
  
  if (ndrugs != nrow(K))
    stop("do.RANKS: adjacency and label matrices do not agree");
  

  K <- Sparsify.matrix(K, k=1);
  K <- kernel(K, a);
  if (p>1)
    K <- p.step.rw.kernel(K, p=p);

  # construction of the matrix of scores. 
  S1  <- matrix(numeric(nclasses * ndrugs), nrow=ndrugs);
  drugnames<-rownames(T);
  classnames<-colnames(T);
  rownames(S1) <- drugnames;
  colnames(S1) <- classnames;
  S10 <- S1;

  # computing scores for each class
  score.name <- as.character(substitute(score));
  
  if (score.name == "KNN.score")
    for (i in 1:nclasses)  {
      ind.pos <- which(T[,i]==1);
      # 1 CV
      res <- multiple.ker.score.cv(K, ind.pos, m=5, p=1, init.seed=0, fun=score, k=k)$av.scores; 
      S1[,i] <- res;
      res <- multiple.ker.score.cv(K, ind.pos, m=5, p=10, init.seed=0, fun=score, k=k)$av.scores;
      S10[,i] <- res;
      cat("Class ", i, " : ", classnames[i], "\n");
    }
  else 
    for (i in 1:nclasses)  {
      ind.pos <- which(T[,i]==1);
      # 1 CV
      res <- multiple.ker.score.cv(K, ind.pos, m=5, p=1, init.seed=0, fun=score)$av.scores; 
      S1[,i] <- res;
      res <- multiple.ker.score.cv(K, ind.pos, m=5, p=10, init.seed=0, fun=score)$av.scores;
      S10[,i] <- res;
      cat("Class ", i, " : ", classnames[i], "\n");
    }  

  # saving scores
  if (score.name == "KNN.score")
    score.name <- paste0(score.name, k);
  
  score.file = paste0("Results/Scores.",score.name,".","p",p,".","a",a,".",data.name,".rda");
  save(S1, S10, file=score.file);
  
  # computing and saving AUC  
  
  AUC.S1 <- AUC.single.over.classes(T, S1);
  AUC.S10 <- AUC.single.over.classes(T, S10);
  
  AUC.file = paste0("Results/AUC.",score.name,".","p",p,".","a",a,".",data.name,".rda");
  #AUC.S10.file = paste0("Results/AUC.",score.name,".","p",p,".","a",a,".","data.name","S10.rda");
  
  save(AUC.S1, AUC.S10, file=AUC.file);
  
  # computing and saving PXR 
  PXR.S1 <- precision.at.multiple.recall.level.over.classes (T, S1, rec.level=recall.levels);
  PXR.S10 <- precision.at.multiple.recall.level.over.classes (T, S10, rec.level=recall.levels);
  
  PXR.file = paste0("Results/PXR.",score.name,".","p",p,".","a",a,".",data.name,".rda");
  
  save(PXR.S1, PXR.S10, file=PXR.file);
  
}













