# Implemntation of Random walk and Random walk with restart
# August 2011

#library(graph);
#library(Rgraphviz);

#source("NetPreProc.R");
# depends NetPreProc (for the method Prob.norm)

library(NetPreProc);
source("sw/ranks.R");

norm1 <- function(x) {
 return(sum(abs(x)));
}



# Function that performs a random Walk on a given graph
# Input:
# W : adjacency matrix of the graph
# ind.positives: indices of the "core" positive examples of the graph.
#                They represent to the indices of W corresponding to the positive examples
# tmax : maximum number of iterations (def: 1000)
# eps : maximum allowed difference between the computed probabilities at the steady state (def. 1e-10)
# norm : if TRUE (def) the adjacency matrix W of the graph is normalized to M = D^-1 * W, otherwise
#        it is assumed that the matrix W is just normalized
# Output:
# a list with three elements:
# - p : the probability at the steady state
# - ind.positives: indices of the "core" positive examples of the graph (it is equal to the same
#                  input parameter
# - n.iter : number of performed iterations
random.walk <- function(W, ind.positives, tmax=1000, eps=1e-10, norm=TRUE) { 
   if (norm) 
     M <-Prob.norm(W) # M = D^-1 * W
   else
     M <- W;
   n <- nrow(M);
   p0 <- p <- numeric(n);
   names(p) <- names(p0) <- rownames(W);
   n.positives <- length(ind.positives);
   if (n.positives == 0)
 	  stop("Random.walk: number of core positives is equal to 0!");
   p0[ind.positives] <- 1/n.positives;
   
   M <- t(M);
   p <- p0;
   for (t in 1:tmax) {
 	 pold <- p;
 	 p <- M %*% pold;
 	 if (norm1(p-pold) < eps) break(); 
   }  
   return(list(p=p, ind.positives=ind.positives, n.iter=t));  
}


# Function that performs a random Walk with restart (RWR) on a given graph
# Input.
# W : adjacency matrix of the graph
# ind.positives: indices of the "core" positive examples of the graph.
#                They represent to the indices of W corresponding to the positive examples
# gamma : restart parameter (def: 0.6)
# tmax : maximum number of iterations (def: 1000)
# eps : maximum allowed difference between the computed probabilities at the steady state
# norm : if TRUE (def) the adjacency matrix W of the graph is normalized to M = D^-1 * W, otherwise
#        it is assumed that the matrix W is just normalized
# Output:
# a list with three elements:
# - p : the probability at the steady state
# - ind.positives: indices of the "core" positive examples of the graph (it is equal to the same
#                  input parameter
# - n.iter : number of performed iterations
RWR <- function(W, ind.positives, gamma=0.6, tmax=1000, eps=1e-10, norm=TRUE) {  
   if (norm) 
     M <-Prob.norm(W) # M = D^-1 * W
   else
     M <- W;
   n <- nrow(M);
   p0 <- p <- numeric(n);
   names(p) <- names(p0) <- rownames(W);
   n.positives <- length(ind.positives);
   if (n.positives == 0)
 	  stop("Random.walk: number of core positives is equal to 0!");
   p0[ind.positives] <- 1/n.positives;
   
   M <- t(M);
   p <- p0;
   for (t in 1:tmax) {
 	 pold <- p;
 	 p <- ((1-gamma) * M %*% pold) + gamma * p0;
 	 if (norm1(p-pold) < eps) break(); 
   }  
   return(list(p=p, ind.positives=ind.positives, n.iter=t));  
}


# Function that implements the Label propagation algorithm of Zhu and Ghahramani
# Input:
# W : adjacency matrix of the graph
# ind.positives: indices of the "core" positive examples of the graph.
#                They represent to the indices of W corresponding to the positive examples
# tmax : maximum number of iterations (def: 1000)
# eps : maximum allowed difference between the computed probabilities at the steady state (def. 1e-5)
# norm : if TRUE (def) the adjacency matrix W of the graph is normalized to M = D^-1 * W, otherwise
#        it is assumed that the matrix W is just normalized
# Output:
# a list with three elements:
# - p : the probability at the steady state
# - ind.positives: indices of the "core" positive examples of the graph (it is equal to the same
#                  input parameter
# - n.iter : number of performed iterations
label.prop <- function(W, ind.positives, tmax=1000, eps=1e-5, norm=TRUE) { 
   if (norm) 
     M <-Prob.norm(W) # M = D^-1 * W
   else
     M <- W;
   n <- nrow(M);
   p <- numeric(n);
   names(p)  <- rownames(W);
   n.positives <- length(ind.positives);
   if (n.positives == 0)
 	  stop("label.prop: number of core positives is equal to 0!");
   p[ind.positives] <- 1;
   
   M <- t(M);
   for (t in 1:tmax) {
 	 pold <- p;
 	 p <- M %*% pold;
	 p[ind.positives] <- 1;
 	 if (norm1(p-pold) < eps) break();  
   }  
   return(list(p=p, ind.positives=ind.positives, n.iter=t));  
}

# Function that implements a simple Guilt By Association (GBA) method for  label ranking based on 
# averaging between the edge weights connecting a node to its positive neighbours
# Input:
# W : adjacency matrix of the graph
# ind.positives: indices of the "core" positive examples of the graph.
#                They represent to the indices of W corresponding to the positive examples
# Output:
# a list with one element:
# - p : score associated to each node
GBAav <- function(W, ind.positives) {
   n <- nrow(W);
   p <- numeric(n);
   names(p)  <- rownames(W);
   n.positives <- length(ind.positives);
   if (n.positives == 0)
 	  stop("simpleGBA: number of core positives is equal to 0!");
   diag(W)<-0;
   p <- apply(W[,ind.positives],1,sum);
   res <- list(p=p);
   return(res);
}

# Function that implements a simple Guilt By Association (GBA) method for  label ranking based on 
# the maximum between the edge weights connecting a node to its positive neighbours
# Input:
# W : adjacency matrix of the graph
# ind.positives: indices of the "core" positive examples of the graph.
#                They represent to the indices of W corresponding to the positive examples
# Output:
# a list with one element:
# - p : score associated to each node
GBAmax <- function(W, ind.positives) {
   n <- nrow(W);
   p <- numeric(n);
   names(p)  <- rownames(W);
   n.positives <- length(ind.positives);
   if (n.positives == 0)
 	  stop("simpleGBAmax: number of core positives is equal to 0!");
   diag(W)<-0;
   p <- apply(W[,ind.positives],1,max);
   res <- list(p=p);
   return(res);
}


######################################################################
## Method to perform a simple cross-validation 

# Function to execute cross-validation with random walk based method
# Input:
# W : adjacency matrix of the graph
#     note that if the optional argument norm=TRUE (def.), the W matrix is normalized, otherwise it
#     is assumed that W is just normalized
# ind.positives: indices of the "core" positive examples of the graph.
#                They represent to the indices of W corresponding to the positive examples
# k : number of folds (def: 5)
# init.seed : initial seed for the random generator. If 0 (default) no initialization is performed
# fun : function. It must be a randow walk method:
#                 - random.walk (default)
#                 - RWR
#                 - label.prop
# ... : optional arguments for the function fun:
# - gamma : restart parameter (def: 0.6) (meaningful only for RWR)
# - tmax : maximum number of iterations (def: 1000)
# - eps : maximum allowed difference between the computed probabilities at the steady state (def. 1e-10)
# Output
# a vector with the the probabilities for each example at the steady state

RW.cv <- function(W, ind.pos, k=5, init.seed=0, fun=random.walk, ...) {
   if (init.seed!=0)
     set.seed(init.seed);
   n <- nrow(W);
   p <- numeric(n);
   names(p) <- rownames(W);
   
   # Realization of the folds. ERA QUI L'ERRORE: prima al posto di 1:n c'era solo n !!!!
   fold <- do.stratified.cv.data(1:n, ind.pos, k=k, seed=init.seed);
   
   # computing scores on the k folds
   for (i in 1:k) {
     x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
     core.pos <- integer(0);
     for (j in 1:k)
       if (j!=i)
	 core.pos <- c(core.pos, fold$fold.positives[[j]]);
     p[x] <- fun(W,core.pos, ...)$p[x];
   }
   return(p);
}


######################################################################
## Method to perform multiple cross-validation

# Function to execute multiple cross-validation with a kernel-based score method.
# It computes the scores by averaging across multiple cross validations
# Input:
# W : adjacency matrix of the graph
#     note that if the optional argument norm=TRUE (def.), the W matrix is normalized, otherwise it
#     is assumed that W is just normalized
# ind.positives: indices of the "core" positive examples of the graph.
#                They represent to the indices of W corresponding to the positive examples
# k : number of folds (def: 5)
# p : number of repeated cross-validations
# init.seed : initial seed for the random generator. If 0 (default) no initialization is performed
# fun : function. It must be a randow walk method:
#                 - random.walk (default)
#                 - RWR
#                 - label.prop
# ... : optional arguments for the function fun:
# - gamma : restart parameter (def: 0.6) (meaningful only for RWR)
# - tmax : maximum number of iterations (def: 1000)
# - eps : maximum allowed difference between the computed probabilities at the steady state (def. 1e-10)
# Output
# a vector with the the probabilities for each example at the steady state averaged across multiple cross-validations
multiple.RW.cv <- function(W, ind.pos, k=5, p=100, init.seed=0, fun=random.walk, ...) {
   n <- nrow(W);
   current.p <- average.p <- numeric(n);  # vector of average scores
   
   for (v in 1:p) {
	 # Realization of the k folds
	 fold <- do.stratified.cv.data(1:n, ind.pos, k=k, seed=v+init.seed);
	 # computing scores on the k folds
	 for (i in 1:k) {
       x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
       core.pos <- integer(0);
       for (j in 1:k)
    	 if (j!=i)
	   core.pos <- c(core.pos, fold$fold.positives[[j]]);
       current.p[x] <- fun(W,core.pos, ...)$p[x]; 
	 }
	 average.p <- average.p + current.p;
   }
   return(average.p/p);
}

