# Scores based on kernels for ranking nodes in a graph 
# January 2011
# Modified August 2011
# Modified October 2011
# Modified December 2011
# Modified January 2012

library(graph);

dyn.load("ranks.so")

#################################
## Graph kernels
#################################

########################################################
setGeneric("rw.kernel", 
                 function(W, a=2) standardGeneric("rw.kernel"));
		 
# Method to compute the random walk kernel (Smola and Kondor, 2003)
# It computes RW = (a-1)I + D^(-1/2) * W * D^(-1/2)
# where I is the identity matrix, W is the weight matrix of an undirected graph, 
# and  D is a diagonal matrix with D_ii = sum_j W_ij
# Input:
# W : a square symmetric matrix with positive values
# a : numeric. It is correlated to the probability of remaining at the same vertex. 
#              Larger a, larger the probability (def. 2)
# Output
# the one-step random walk kernel matrix
setMethod("rw.kernel", signature(W="matrix"),
  function(W, a=2) {
   n <- nrow(W);
   if (n != ncol(W))
     stop("first arg must be a square matrix");
   if (a < 1)
     stop("coefficient a must be larger or equal than 1");
   # dyn.load("ranks.so");
   names.var <- rownames(W);
   # computing (a-1)I
   I <- diag(x=a-1, nrow=n);
   # computing D^(-1/2) * W * D^(-1/2)
   diag.D <- apply(W,1,sum);
   diag.D[diag.D==0] <- Inf;
   inv.sqrt.diag.D <- 1/sqrt(diag.D);
   W <- .C("norm_lapl_graph", as.double(W), as.double(inv.sqrt.diag.D), as.integer(n))[[1]];
   W <- I + W;
   rownames(W) <- colnames(W) <- names.var;
   return(W);
})


# Method to compute the random walk kernel (Smola and Kondor, 2003)
# Input:
# W : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# a : numeric. It is correlated to the probability of remaining at the same vertex. 
#              Larger a, larger the probability (def. 2)
# Output
# the one-step random walk kernel matrix
setMethod("rw.kernel", signature(W="graph"),
function(W, a=2) {
  W <- as(W, "matrix");
  return(rw.kernel(W, a));
})


##################################################################
setGeneric("p.step.rw.kernel", 
                 function(RW, p=2) standardGeneric("p.step.rw.kernel"));

# Method to compute a p-step random walk kernel.
# It computes a p-step random walk kernel RW^p
# Input:
# RW : a square symmetric matrix representing a one-step random walk kernel
# p : integer. Number of steps (def: p=2)
# Output
# the p-step random walk kernel matrix
setMethod("p.step.rw.kernel", signature(RW="matrix"),
  function(RW, p=2) {
	if (nrow(RW) != ncol(RW))
      stop("first arg must be a square matrix");
	if (p < 2)
      stop("the number of steps a must be larger or equal than 2");
	Res <- RW;
	for (i in 2:p)
      Res <- Res %*% RW;
	return(Res);
})

# Method to compute a p-step random walk kernel.
# It computes a p-step random walk kernel RW^p
# Input:
# RW : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# p : integer. Number of steps (def: p=2)
# Output
# the p-step random walk kernel matrix
setMethod("p.step.rw.kernel", signature(RW="graph"),
  function(RW, p=2) {
	W <- rw.kernel(RW);
	return(p.step.rw.kernel(W,p));
})

		 
# Function to compute a linear  kernel from a feature matrix 
# It computes K = W  * t(W), where W is a feature matrix
# Input:
# W : a feature matrix, Rows are examples and columns are features        
# Output
# a linear kernel matrix 
linear.kernel <- function(W) {
   return(W %*% t(W));
}

# Function to compute a gaussian  kernel from a feature matrix 
# Input:
# W : a feature matrix, Rows are examples and columns are features      
# sigma : a real value representing the sigma parameter of the gaussian  
# Output
# a gaussian kernel matrix 
gaussian.kernel <- function(W, sigma=1) {
   n <- nrow(W);
   m <- ncol(W);
   nn <- rownames(W);
   K <- matrix(numeric(n*n), nrow=n);
   K <- .C("gaussian_kernel", as.double(K), as.double(W), as.double(sigma), as.integer(n), as.integer(m))[[1]];
   K <- matrix(K, nrow=n);
   rownames(K) <- colnames(K) <- nn;
   return(K);
}

# Function to compute a laplacian  kernel from a feature matrix 
# Input:
# W : a feature matrix, Rows are examples and columns are features      
# sigma : a real value representing the sigma parameter of the laplacian  
# Output
# a laplacian kernel matrix 
laplacian.kernel <- function(W, sigma=1) {
   n <- nrow(W);
   m <- ncol(W);
   nn <- rownames(W);
   K <- matrix(numeric(n*n), nrow=n);
   K <- .C("laplacian_kernel", as.double(K), as.double(W), as.double(sigma), as.integer(n), as.integer(m))[[1]];
   K <- matrix(K, nrow=n);
   rownames(K) <- colnames(K) <- nn;
   return(K);
}

# Function to compute a Cauchy  kernel from a feature matrix 
# Input:
# W : a feature matrix, Rows are examples and columns are features      
# sigma : a real value representing the sigma parameter of the cauchy kernel 
# Output
#  a Cauchy  kernel matrix 
cauchy.kernel <- function(W, sigma=1) {
   n <- nrow(W);
   m <- ncol(W);
   nn <- rownames(W);
   K <- matrix(numeric(n*n), nrow=n);
   K <- .C("cauchy_kernel", as.double(K), as.double(W), as.double(sigma), as.integer(n), as.integer(m))[[1]];
   K <- matrix(K, nrow=n);
   rownames(K) <- colnames(K) <- nn;
   return(K);
}

# Function to compute an inverse multiquadric  kernel from a feature matrix 
# Input:
# W : a feature matrix, Rows are examples and columns are features      
# v : constant factor (def. 1). v must be larger than 0.
# Output
#  an inverse multiquadric  kernel matrix 
inv.multiquadric.kernel <- function(W, v=1) {
   n <- nrow(W);
   m <- ncol(W);
   nn <- rownames(W);
   K <- matrix(numeric(n*n), nrow=n);
   K <- .C("inv_mq_kernel", as.double(K), as.double(W), as.double(v), as.integer(n), as.integer(m))[[1]];
   K <- matrix(K, nrow=n);
   rownames(K) <- colnames(K) <- nn;
   return(K);
}


# Function to compute a polynomial  kernel from a feature matrix 
# Input:
# W : a feature matrix, Rows are examples and columns are features      
# degree : integer corresponding to a degree of the polynomial (def.: 2)
# scale : scaling factor. Double larger than 0. If scale=-1 (def) scale is set to 1/ncol(W);
# v : constant factor (def. 0)
# Output
# a gaussian kernel matrix 
poly.kernel <- function(W, degree=2, scale=-1, v=0) {
   n <- nrow(W);
   m <- ncol(W);
   nn <- rownames(W);
   if (scale == -1)
     scale <- 1/as.double(m);
   K <- matrix(numeric(n*n), nrow=n);
   K <- .C("poly_kernel", as.double(K), as.double(W), as.integer(degree), as.double(scale), as.double(v), as.integer(n), as.integer(m))[[1]];
   K <- matrix(K, nrow=n);
   rownames(K) <- colnames(K) <- nn;
   return(K);
}




################################################################
## Score functions 
## N.B. Per questi metodi i primi 2 argomenti risultano invertiti rispetto alla precedente versione
################################################################

# eav.score  Empirical Average Score 

########################################################
setGeneric("single.eav.score", 
                 function(RW, x, x.pos, auto=FALSE) standardGeneric("single.eav.score"));

# Method to compute the Empirical Average score for a single vertex
# score(x) = - K(x,x) + 2/|x.pos| * sum_{x_i \in x.pos} K(x,x_i)
# Input:
# RW : matrix. It must be a kernel matrix or a symmetric matrix expressing the similarity
#              between pairs of nodes.
# x : integer. Index corresponding to the element of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the component - K(x,x) is computed, otherwise is discarded (default)
# Output
# the eav score of the element x
setMethod("single.eav.score", signature(RW="matrix"),
  function(RW, x, x.pos, auto=FALSE) {
   if (nrow(RW) != ncol(RW))
     stop("second arg must be a square matrix");
   if (auto)
     score <- -RW[x,x] + (2/length(x.pos)) * sum(RW[x,x.pos])   
   else
     score <- (2/length(x.pos)) * sum(RW[x,x.pos]); 
   return(score);
})


# Method to compute the Empirical Average score for a single vertex
# Input:
# RW : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# x : integer. Index corresponding to the element of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the component - K(x,x) is computed, otherwise is discarded (default)
# Output
# the eav score of the element x
setMethod("single.eav.score", signature(RW="graph"),
  function(RW, x, x.pos, auto=FALSE) {
      RW <- as(RW, "matrix");
      return(single.eav.score(RW, x, x.pos, auto));
})


########################################################
setGeneric("eav.score", 
                 function(RW, x, x.pos, auto=FALSE) standardGeneric("eav.score"));

# Method to compute the Empirical Average score for a set of vertices
# score(x) = - K(x,x) + 2/|x.pos| * sum_{x_i \in x.pos} K(x,x_i)
# Input:
# x : vector of integer. Indices corresponding to the elements of the RW matrix for which the score must be computed
# RW : matrix. It must be a kernel matrix or a symmetric matrix expressing the similarity
#              between pairs of nodes.
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the component - K(x,x) is computed, otherwise is discarded (default)
# Output
# vector of the eav scores of the elements x. The names of the vector correspond to the indices x
setMethod("eav.score", signature(RW="matrix"),
  function(RW, x, x.pos, auto=FALSE) {
	if (nrow(RW) != ncol(RW))
      stop("second arg must be a square matrix");
	if (auto)
      score <- -diag(RW)[x] + (2/length(x.pos)) * apply(RW[x,x.pos],1,sum) 
	else
      score <- (2/length(x.pos)) * apply(RW[x,x.pos],1,sum);
	names(score)<-x;  
	return(score);
})


# Method to compute the Empirical Average score for a set of vertices
# Input:
# x : vector of integer. Indices corresponding to the elements of the RW matrix for which the score must be computed
# RW : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the component - K(x,x) is computed, otherwise is discarded (default)
# Output
# vector of the eav scores of the elements x. The names of the vector correspond to the indices x
setMethod("eav.score", signature(RW="graph"),
  function(RW, x, x.pos, auto=FALSE) {
     RW <- as(RW, "matrix");
     return(single.eav.score(RW, x, x.pos, auto));
})



setGeneric("single.NN.score", 
                 function(RW, x, x.pos, auto=FALSE) standardGeneric("single.NN.score"));
				 
# Method to compute the Kernel NN score for a single vertex
# score(x) = - min__{x_i \in x.pos} ( K(x,x) + K(x_i,x_i) -2 K(x,x_i)
# Input:
# RW : matrix. It must be a kernel matrix or a symmetric matrix expressing the similarity
# x : integer. Index corresponding to the element of the RW matrix for which the score must be computed
#              between pairs of nodes.
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded (default)
# Output
# the NN score of the element x
setMethod("single.NN.score", signature(RW="matrix"),
  function(RW, x, x.pos, auto=FALSE) {
	if (nrow(RW) != ncol(RW))
      stop("second arg must be a square matrix");
	n <- length(x.pos);
	if (auto) {
      scores <- rep(RW[x,x],n) + diag(RW)[x.pos] - 2 * RW[x,x.pos];
      names(scores) <- x;
      score <-  - min(scores);           
	} else {
      scores <- RW[x,x.pos];
      names(scores) <- x;
      score <- max(scores);
	}
	return(score);
})


# Method to compute the Kernel NN score for a single vertex
# Input:
# RW : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# x : integer. Index corresponding to the element of the RW matrix for which the score must be computed
#              between pairs of nodes.
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded (default)
# Output
# the NN score of the element x
setMethod("single.NN.score", signature(RW="graph"),
  function(RW, x, x.pos, auto=FALSE) {
     RW <- as(RW, "matrix");
     return(single.NN.score(RW, x, x.pos, auto));
})


setGeneric("NN.score", 
                 function(RW, x, x.pos, auto=FALSE) standardGeneric("NN.score"));

# Method to compute the Kernel NN score for a set of vertices
# score(x) = - min__{x_i \in x.pos} ( K(x,x) + K(x_i,x_i) -2 K(x,x_i)
# Input:
# RW : matrix. It must be a kernel matrix or a symmetric matrix expressing the similarity
#              between pairs of nodes.
# x : vector of integer. Indices corresponding to the elements of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded (default)
# Output
# the vector of the NN scores of the elements x. The names of the vector correspond to the indices x
setMethod("NN.score", signature(RW="matrix"),
  function(RW, x, x.pos, auto=FALSE) {
   if (nrow(RW) != ncol(RW))
     stop("second arg must be a square matrix");
   n <- length(x.pos);
   if (auto) {
     scores <- matrix(rep(diag(RW)[x],n), ncol=n) +
     matrix(rep(diag(RW)[x.pos], length(x)), nrow=length(x), byrow=T) - 2 * RW[x,x.pos];
     score <- - apply(scores,1,min);
	 names(score)<-x;           
   } else {
     scores <- RW[x,x.pos];
     score <- apply(scores,1,max); 
	 names(score)<-x;
   }
   return(score);
})

# Method to compute the Kernel NN score for a set of vertices
# Input:
# RW : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# x : vector of integer. Indices corresponding to the elements of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded (default)
# Output
# the vector of the NN scores of the elements x. The names of the vector correspond to the indices x
setMethod("NN.score", signature(RW="graph"),
  function(RW, x, x.pos, auto=FALSE) {
     RW <- as(RW, "matrix");
     return(NN.score(RW, x, x.pos, auto));
})


setGeneric("single.KNN.score", 
                 function(RW, x, x.pos, k=3, auto=FALSE) standardGeneric("single.KNN.score"));

# Method to compute the Kernel KNN score for a single vertex
# score(x) = - sum_{k nearest x_i \in x.pos} ( K(x,x) + K(x_i,x_i) -2 K(x,x_i)
# Input:
# RW : matrix. It must be a kernel matrix or a symmetric matrix expressing the similarity
#              between pairs of nodes.
# x : integer. Index corresponding to the element of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded (default)
# k : integer. Number of the k nearest neighbours to be considered
# Output
# the KNN score of the element x
setMethod("single.KNN.score", signature(RW="matrix"),
  function(RW, x, x.pos, k=3, auto=FALSE) {
	if (nrow(RW) != ncol(RW))
      stop("second arg must be a square matrix");
	n <- length(x.pos);
	if (k > n) {
      k <- n;
      warn.message <- paste("k must be lower or equal to the number of positive examples: k set to", n);
      warning(warn.message);
	}     
	if (auto) {
      scores <- rep(RW[x,x],n) + diag(RW)[x.pos] - 2 * RW[x,x.pos];
      names(scores) <- x;
      scores <- sort(scores)[1:k];
      score <-  - sum(scores);           
	} else {
      scores <- RW[x,x.pos];
      names(scores) <- x;
      scores <- sort(scores, decreasing=TRUE)[1:k];
      score <- sum(scores);
	}
	return(score/k);
})

# Method to compute the Kernel KNN score for a single vertex
# Input:
# RW : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# x : integer. Index corresponding to the element of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded (default)
# k : integer. Number of the k nearest neighbours to be considered
# Output
# the KNN score of the element x
setMethod("single.KNN.score", signature(RW="graph"),
  function(RW, x, x.pos, k=3, auto=FALSE) {
    RW <- as(RW, "matrix");
    return(single.KNN.score(RW, x, x.pos, k, auto));
})


setGeneric("KNN.score", 
                 function(RW, x, x.pos, k=3, auto=FALSE) standardGeneric("KNN.score"));

# Method to compute the Kernel KNN score for a set of vertices. 
# score(x) = - (1/k) |sum_{k nearest x_i \in x.pos} ( K(x,x) + K(x_i,x_i) -2 K(x,x_i)
# Input:
# RW : matrix. It must be a kernel matrix or a symmetric matrix expressing the similarity
#              between pairs of nodes.
# x : vector of integer. Indices correspond to the elements of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded 
# (default). The option TRUE is at the moment significantly slower.
# k : integer. Number of the k nearest neighbours to be considered (def. 3)
# Output
# the KNN scores of the elements x
setMethod("KNN.score", signature(RW="matrix"),
  function(RW, x, x.pos, k=3, auto=FALSE) {
   if (length(x.pos)==1)
     stop("The number of positives must be larger than 1. This bug must be fixed");
   if (nrow(RW) != ncol(RW))
     stop("second arg must be a square matrix");
   # dyn.load("ranks.so");
   n <- length(x.pos);
   if (k==1)
     stop("k must be larger than 1. Please, use NN.score instead.");
   if (k > n) {
     k <- n;
     warn.message <- paste("k must be lower or equal to the number of positive examples: k set to", n);
     warning(warn.message);
   }     
   if (auto) {	 
	 scores <- matrix(rep(diag(RW)[x],n), ncol=n) +
     matrix(rep(diag(RW)[x.pos], length(x)), nrow=length(x), byrow=T) - 2 * RW[x,x.pos];
     score <- apply(scores,1,sort); # results are "per column"
     score <- score[1:k,]; 
     score <- - apply(score,2,sum);
     names(score) <- x;                   
   } else {
     scores <- RW[x,x.pos];
	 y <- numeric(k);
	 score = matrix(numeric(length(x)*k), ncol=k);
	 for (j in 1:length(x))
        score[j,] <- .C("select_top", as.double(scores[j,]), as.double(y), as.integer(n), as.integer(k))[[2]]; 
     score <- apply(score,1,sum);  # results are "per rows"
     names(score) <- x; 
   }
   return(score/k);
}) 


# Method to compute the Kernel KNN score for a set of vertices. 
# Input:
# RW : an object of the virtual class graph (hence including objects of class graphAM  and graphNEL from the package graph).
# x : vector of integer. Indices correspond to the elements of the RW matrix for which the score must be computed
# x.pos : vector of integer. Indices of the positive elements of the RW matrix
# auto : boolean. If TRUE the components K(x,x) + K(x_i,x_i) are computed, otherwise are discarded 
# (default). The option TRUE is at the moment significantly slower.
# k : integer. Number of the k nearest neighbours to be considered (def. 3)
# Output
# the KNN scores of the elements x
setMethod("KNN.score", signature(RW="graph"),
  function(RW, x, x.pos, k=3, auto=FALSE) {
    RW <- as(RW, "matrix");
    return(KNN.score(RW, x, x.pos, k, auto));
}) 


# Function to compute the labels of multiple classes from the corresponding scores
# Input:
# S : matrix. Matrix of scores: rows represent examples, columns classes
# thresh : numeric vector. Vector of the thesholds for each class.
# Output:
# a binary matrix with the labels of the predictions: rows represent examples, columns classes.
# Element L[i,j] is the label of example i w.r.t. class j.  L[i,j]=1 if i belongs to j, 0 otherwise.
Multiple.labels.from.scores <- function(S, thresh) {
   pred.label <- function(x,thresh) {
	  ii<<-ii+1;
	  res <- ifelse(x>=thresh[ii], 1, 0);
	  return(res);
   }
   ii<<-0;
   L <- apply(S,2, pred.label, thresh=thresh); 
   rm(ii,pos=1);
   rownames(L) <- rownames(S);
   colnames(L) <- colnames(S);
   return(L);
}

# Function to compute the labels of a single class from the corresponding scores
# Input:
# scores : numeric. Vector of scores: each element correspond to the score of an example
# thresh : real value. Threshold for the class.
# Output:
# Labels of the predictions: a vector with 0 or 1 value
# The label res[i]=1 if scores[i]>thresh, otherwise res[i]=0
labels.from.scores <- function(scores, thresh) {
   res <- ifelse(scores>=thresh, 1, 0);
   names(res) <- names(scores);
   return(res);
}

# da qui in poi sono mantenute le funzioni, ma si deve poi vedere se abbia senso usare dei metodi che utilizzino sia
# matrici sia oggetti di classe graph

#################################
## Functions for cross validation
#################################


######################################################################
## Method to perform a simple cross-validation with ranking

# Function to perform a simple cross-validation with a kernel-based score method
# Input:
# RW : matrix. It could be a kerenl matrix or the adjacency matrix of a graph
# ind.pos: indices of the positive examples. They are the indices the row of RW corresponding to
#          positive examples.
# kk : number of folds (def: 5)
# init.seed : initial seed for the random generator. If 0 (default) no initialization is performed
# fun : function. It must be a kernel-based score method:
#                 - KNN.score (default)
#                 - NN.score
#                 - eav.score
# ... : optional arguments for the function fun
# Output
# a vector with the scores computed on each example

ker.score.cv <- function(RW, ind.pos, kk=5, init.seed=0, fun=KNN.score, ...) {
   if (init.seed!=0)
     set.seed(init.seed);
   n <- nrow(RW);
   scores <- numeric(n);
   names(scores) <- rownames(RW);
   
   # Realization of the folds
   fold <- do.stratified.cv.data(1:n, ind.pos, k=kk, seed=init.seed);
   
   # computing scores on the k folds
   for (i in 1:kk) {
     x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
     core.pos <- integer(0);
     for (j in 1:kk)
       if (j!=i)
	 core.pos <- c(core.pos, fold$fold.positives[[j]]);
     scores[x] <- fun(RW, x, core.pos, ...);
   }
   return(scores);
}


#####################################################################
#### Functions to perform multiple-cross validations and computing average scores
#####################################################################


# Function to execute multiple cross-validation with a kernel-based score method for ranking
# It computes the scores by averaging across multiple cross validations
# Input:
# RW : matrix. Random walk matrix
# ind.pos: indices of the positive examples. They are the indices the row of RW corresponding to
#          positive examples.
# m : number of folds for each cross-validation
# p : number of repeated cross-validations
# init.seed : initial seed for the random generator (def: 0)
# fun : function. It must be a kernel-based score method:
#                 - KNN.score (default)
#                 - NN.score 
#                 - eav.score
# ... : optional arguments for the function fun
# Output
# a list with two components:
# - av.scores : a vector with the average scores across multiple cross-validations.
#               Elements of the vector av.scores correspond to the rows of RW
# - pos.scores : a vector with the scores of positive elements collected at each iteration
multiple.ker.score.cv <-  function(RW, ind.pos, m=5, p=100, init.seed=0, fun=KNN.score, ...) {
   n <- nrow(RW);
   current.scores <- av.scores <- numeric(n);  # vector of average scores
   pos.scores <- numeric();  # vector collecting the scores of positive examples
   
   for (v in 1:p) {
	 # Realization of the m folds
	 fold <- do.stratified.cv.data(1:n, ind.pos, k=m, seed=v+init.seed);
	 # computing scores on the m folds
	 for (i in 1:m) {
       x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
       core.pos <- integer(0);
       for (j in 1:m)
    	 if (j!=i)
	   core.pos <- c(core.pos, fold$fold.positives[[j]]);
       current.scores[x] <- fun(RW,x,core.pos, ...);	   
	 }
	 av.scores <- av.scores + current.scores;
	 pos.scores <- c(pos.scores,current.scores[ind.pos]);
   }
   return(list(av.scores=av.scores/p, pos.scores=pos.scores));
}





# Function to find the optimal quantile alpha and corresponding threshold by  cross-validation with a kernel-based 
# score method. The optimality is computed with respect to a specific metric( def: F-score).
# Input:
# RW : matrix. Random walk matrix
# ind.pos: indices of the positive examples. They are the indices the row of RW corresponding to
#          positive examples of the training set.
# ind.non.pos: indices of the non positive examples. They are the indices the row of RW corresponding to
#          non positive examples  of the training set.
# m : number of folds (def: 5)
# alpha : vector of the quantiles to be tested
# init.seed : initial seed for the random generator. If 0 (def) no initialization is performed
# opt.fun : function. Function implementing the metric to choice the optimal threshold.
#           The F-score (compute.F) is the default.
#           Available functions:
#                 - compute.F (default) : F-score
#                 - compute.acc  : accuracy
#           N.B.: any function having two arguments representing the vector of predicted and true labels
#           can be in principle used.
# fun : function. It must be a kernel-based score method:
#                 - KNN.score (default)
#                 - NN.score
#                 - eav.score
# ... : optional arguments for the function fun
# Output
# a list with 3 elements:
# alpha : quantile corresponding to the best F-score
# thresh : threshold corresponding to the best F-score
# pos.scores : scores of the positive elements computed through CV

find.optimal.thresh.cv <- function(RW, ind.pos, ind.non.pos, m=5, 
   alpha=seq(from=0.05, to=0.6, by=0.05), init.seed=0, opt.fun=compute.F, fun=KNN.score, ...) {
   
   if (init.seed !=0)
     set.seed(init.seed);
   ind.all <- c(ind.pos, ind.non.pos);
   n=length(ind.all);
   RW <- RW[ind.all, ind.all];
   ind.pos <- 1:length(ind.pos);
   ind.non.pos <- (length(ind.pos)+1):n;
   av.scores <- numeric(n);  # vector of average scores
   pos.scores <- numeric();  # vector collecting the scores of positive examples cumulated across the folds
   fold <- do.stratified.cv.data(1:n, ind.pos, k=m, seed=init.seed);
   for (i in 1:m) {
     x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
     core.pos <- integer(0);
     for (j in 1:m)
       if (j!=i)
	     core.pos <- c(core.pos, fold$fold.positives[[j]]);
     av.scores[x] <- fun(RW,x,core.pos, ...);	   
   }
   pos.scores <- av.scores[ind.pos];
   
   opt.F = -1;
   opt.alpha=0;
   labels <- c(rep(1, times=length(ind.pos)), rep(0, times=length(ind.non.pos)));
   for (i in 1:length(alpha)) {
     threshold <- selection.test(pos.scores, av.scores, ind.positives=ind.pos, alpha=alpha[i])$thresh;
     pred <- ifelse(av.scores>=threshold, 1, 0);
	 F <- opt.fun(pred,labels);    
	 if (F > opt.F) {
	   opt.F <- F;
	   opt.alpha <- alpha[i];
	   opt.thresh <- threshold;
	 }
   }
   return(list(alpha=opt.alpha, thresh=opt.thresh, pos.scores=pos.scores));
}



# Function to execute multiple cross-validation with a kernel-based score method and to find the optimal
# threshold for a given class by internal cross-validation. 
# Scores are computed by averaging across multiple external cross-validations.
# The optimal quantile and corresponding threshold  are selected by internal cross-validation using a 
# specific metric (def: F-score).
# Input:
# RW : matrix. Random walk matrix
# ind.pos: indices of the positive examples. They are the indices the row of RW corresponding to
#          positive examples.
# m : number of folds for each cross-validation (the number applies to both external and internal CV)
# p : number of repeated external cross-validations
# alpha : vector of the quantiles to be tested
# init.seed : initial seed for the random generator (def: 0)
# fun : function. It must be a kernel-based score method:
#                 - KNN.score (default)
#                 - NN.score 
#                 - eav.score
# ... : optional arguments for the function fun
# Output
# a list with three components:
# - av.scores : a vector with the average scores across multiple cross-validations.
#               Elements of the vector av.scores correspond to the rows of RW
# - opt.alpha : the optimal alpha for the class
# - opt.thresh : the optimal threshold for the class
multiple.ker.score.thresh.cv <- function(RW, ind.pos, m=5, p=100, alpha=seq(from=0.05, to=0.6, by=0.05), init.seed=0, fun=KNN.score, ...) {
   set.seed(init.seed);
   n <- nrow(RW);
   current.scores <- av.scores <- numeric(n);  # vector of average scores
   opt.alpha = 0;
   opt.thresh = 0;
   for (v in 1:p) {
	 # Realization of the m folds
	 fold <- do.stratified.cv.data(1:n, ind.pos, k=m, seed=v+init.seed);
	 # computing scores on the m folds
	 for (i in 1:m) {
       x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
       core.pos <- core.non.pos <- integer(0);
       for (j in 1:m)
    	 if (j!=i) {
	       core.pos <- c(core.pos, fold$fold.positives[[j]]);
	       core.non.pos <- c(core.non.pos, fold$fold.non.positives[[j]]);
	   }
       current.scores[x] <- fun(RW,x,core.pos, ...);
	   res <- find.optimal.thresh.cv(RW, core.pos, core.non.pos, 
	                                m=m, alpha=alpha, init.seed=init.seed, fun=fun, ...);
	   opt.alpha <- opt.alpha + res$alpha;	
	   opt.thresh <- opt.thresh + res$thresh;   	   
	 }
	 av.scores <- av.scores + current.scores;
   }
   return(list(av.scores=av.scores/p, opt.alpha=opt.alpha/(m*p), opt.thresh=opt.thresh/(m*p)));
}


# Function to classify labels according to an hold-out procedure with a kernel-based score method.
# The optimal threshold for a given class is obtained by (possibly multiple) internal cross-validation. 
# Scores of the held-out nodes are computed and thresholds computed on the training set 
# by cross-validation are then used to classify the held-out nodes
# The optimal quantile and corresponding threshold  are selected by internal cross-validation using the F-score as metrics
# Input:
# RW : matrix. Random walk matrix
# ind.pos: indices of the positive examples of the training set. They are the indices the row of RW corresponding to
#          positive examples of the training set
# ind.test : indices of the examples of the test set. They are the indices the row of RW corresponding to
#           examples of the test set
# m : number of folds for each cross-validation 
# p : number of repeated  cross-validations on the training set
# alpha : vector of the quantiles to be tested
# init.seed : initial seed for the random generator (def: 0)
# opt.fun : function. Function implementing the metric to choice the optimal threshold.
#           The F-score (compute.F) is the default.
#           Available functions:
#                 - compute.F (default) : F-score
#                 - compute.acc  : accuracy
#           N.B.: any function having two arguments representing the vector of predicted and true labels
#           can be in principle used.
# fun : function. It must be a kernel-based score method:
#                 - KNN.score (default)
#                 - NN.score 
#                 - eav.score
# ... : optional arguments for the function fun
# Output
# a list with four components:
# - labels : vector of the predicted labels for the test set(1 represent positive, 0 negative)
# - scores : a vector with the  scores computed on the test set.
#               Elements of the vector av.scores correspond to ind.test rows of RW
# - opt.alpha : the optimal alpha for the class
# - opt.thresh : the optimal threshold for the class
ker.score.classifier.holdout <- function(RW, ind.pos, ind.test, m=5, p=100, alpha=seq(from=0.05, to=0.6, by=0.05), init.seed=0, opt.fun=compute.F, fun=KNN.score, ...) {
   set.seed(init.seed);
   n <- nrow(RW);
   ind.train <- (1:n)[-ind.test];
   dd <- intersect(ind.pos, ind.test);
   if (length(dd)>0)
     stop("ker.score.classifier.holdout: conflicting indices between training and test set");
   opt.alpha = 0;
   opt.thresh = 0;
   for (v in 1:p) {
	 # Realization of the m folds
	 fold <- do.stratified.cv.data(ind.train, ind.pos, k=m, seed=v+init.seed);
	 # computing thresholds on the m folds
	 for (i in 1:m) {
       x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
       core.pos <- core.non.pos <- integer(0);
       for (j in 1:m)
    	 if (j!=i) {
	       core.pos <- c(core.pos, fold$fold.positives[[j]]);
	       core.non.pos <- c(core.non.pos, fold$fold.non.positives[[j]]);
	   }
	   res <- find.optimal.thresh.cv(RW, core.pos, core.non.pos, 
	                                m=m, alpha=alpha, init.seed=init.seed, opt.fun=opt.fun, fun=fun, ...);
	   opt.alpha <- opt.alpha + res$alpha;	
	   opt.thresh <- opt.thresh + res$thresh;   	   
	 }
   }
   opt.alpha <- opt.alpha/(m*p); 
   opt.thresh <- opt.thresh/(m*p);
   scores <- fun(ind.test,RW,ind.pos, ...);
   labels <- labels.from.scores(scores, opt.thresh);
   names(labels) <- names(scores) <- names(ind.test);
   return(list(labels=labels, scores=scores, opt.alpha=opt.alpha, opt.thresh=opt.thresh));
}




# Function to classify labels accroding to external cross-validation procedure with a kernel-based score method and to find the optimal
# threshold for a given class by internal cross-validation. 
# Scores are computed by averaging across (possibly) multiple external cross-validations.
# The optimal quantile and corresponding threshold  are selected by internal cross-validation using the F-score as metrics
# Input:
# RW : matrix. Random walk matrix
# ind.pos: indices of the positive examples. They are the indices the row of RW corresponding to
#          positive examples.
# m : number of folds for each cross-validation (the number applies to both external and internal CV)
# p : number of repeated external cross-validations
# alpha : vector of the quantiles to be tested
# init.seed : initial seed for the random generator (def: 0)
# opt.fun : function. Function implementing the metric to choice the optimal threshold.
#           The F-score (compute.F) is the default.
#           Available functions:
#                 - compute.F (default) : F-score
#                 - compute.acc  : accuracy
#           N.B.: any function having two arguments representing the vector of predicted and true labels
#           can be in principle used.
# fun : function. It must be a kernel-based score method:
#                 - KNN.score (default)
#                 - NN.score 
#                 - eav.score
# ... : optional arguments for the function fun
# Output
# a list with four components:
# - labels : vector of the predicted labels (1 represent positive, 0 negative)
# - av.scores : a vector with the average scores across multiple cross-validations.
#               Elements of the vector av.scores correspond to the rows of RW
# - opt.alpha : the optimal alpha for the class
# - opt.thresh : the optimal threshold for the class
ker.score.classifier.cv <- function(RW, ind.pos, m=5, p=100, alpha=seq(from=0.05, to=0.6, by=0.05), init.seed=0, opt.fun=compute.F, fun=KNN.score, ...) {
   set.seed(init.seed);
   n <- nrow(RW);
   current.scores <- av.scores <- numeric(n);  # vector of average scores
   opt.alpha = 0;
   opt.thresh = 0;
   for (v in 1:p) {
	 # Realization of the m folds
	 fold <- do.stratified.cv.data(n, ind.pos, k=m, seed=v+init.seed);
	 # computing scores on the m folds
	 for (i in 1:m) {
       x <- c(fold$fold.positives[[i]], fold$fold.non.positives[[i]]);
       core.pos <- core.non.pos <- integer(0);
       for (j in 1:m)
    	 if (j!=i) {
	       core.pos <- c(core.pos, fold$fold.positives[[j]]);
	       core.non.pos <- c(core.non.pos, fold$fold.non.positives[[j]]);
	   }
       current.scores[x] <- fun(x,RW,core.pos, ...);
	   res <- find.optimal.thresh.cv(RW, core.pos, core.non.pos, 
	                                m=m, alpha=alpha, init.seed=init.seed, opt.fun=opt.fun, fun=fun, ...);
	   opt.alpha <- opt.alpha + res$alpha;	
	   opt.thresh <- opt.thresh + res$thresh;   	   
	 }
	 av.scores <- av.scores + current.scores;
   }
   av.scores <- av.scores/p;
   opt.alpha <- opt.alpha/(m*p); 
   opt.thresh <- opt.thresh/(m*p);
   labels <- labels.from.scores(av.scores, opt.thresh)
   return(list(labels=labels, av.scores=av.scores, opt.alpha=opt.alpha, opt.thresh=opt.thresh));
}





###########################################################################
## Non parameteric selection test based on multiple cross-validations
###########################################################################

# Non parametric test to select the most significant unlabeled examples.
# Input:
# pos.scores : vector with scores of positive examples. It is returned from multiple.ker.score.cv.
# av.scores : a vector with the average scores computed by multiple.ker.score.cv.   
#             It may be a named vector. If not, the names attributes corresponding to the indices of the vector are added.
# ind.positives: indices of the positive examples. They are the indices of av.scores corresponding to
#          positive examples.
# alpha : significance level (def. 0.05)
# thresh.pos : only values larger than thresh.pos are retained in pos.scores (def.: 0)
# Output
# a list with 5 components:
# - selected : a named vector with the components of av.scores selected by the test
# - selected.labeled : a named vector with the labeled components of av.scores selected by the test
# - selected.unlabeled : a named vector with the unlabeled components of av.scores selected by the test
# - thresh : the score threshold selected by the test
# - alpha : significance level (the same value of the input)
selection.test <- function(pos.scores, av.scores, ind.positives, alpha=0.05, thresh.pos=0) {
  
  n <- length(av.scores);
  pos.scores <- pos.scores[pos.scores > thresh.pos];
  if (is.null(names(av.scores)))
    names(av.scores) <- 1:n; 
  av.scores.labeled <- av.scores[ind.positives];
  av.scores.unlabeled <- av.scores[-ind.positives];
  thresh <- quantile(pos.scores,alpha);
  selected <- av.scores[av.scores>=thresh];
  selected.labeled <- av.scores.labeled[av.scores.labeled>=thresh];
  selected.unlabeled <- av.scores.unlabeled[av.scores.unlabeled>=thresh];
  
  return(list(selected=selected, selected.labeled=selected.labeled,
              selected.unlabeled=selected.unlabeled, thresh=thresh, alpha=alpha));
}


############# Utility functions ############################# 


######################################################################
# Function to generate data for the stratified cross-validation.
# Input:
# examples : indices of the examples (a vector of integer)
# positives: vector of integer. Indices of the positive examples. The indices refer to row number of
#            the data matrix.
# k : number of folds (def = 5)
# seed : seed of the random generator (def=1). If is set to 0 no initiazitation is performed
# Ouptut:
# a list with 2 two components
#   - fold.non.positives : a list with k components. Each component is a vector with the indices of the non positive elements of the fold
#   - fold.positives : a list with k components. Each component is a vector with the indices of the positive elements of the fold
# N.B.: in both elements indices refer to row numbers of the data matrix	 
do.stratified.cv.data <- function(examples, positives, k=5, seed=1) {
  if (seed!=0)
    set.seed(seed);
  fold.non.positives <- fold.positives <- list();
  for (i in 1:k) {
    fold.non.positives[[i]] <- integer(0);
    fold.positives[[i]] <- integer(0);
  }
  # examples <- 1:n;
  non.positives <- setdiff(examples,positives);
  # non.positives <- examples[-positives];
  non.positives <- sample(non.positives);
  positives <- sample(positives);
  n.positives <- length(positives);
  resto.positives <- n.positives%%k;
  n.pos.per.fold <- (n.positives - resto.positives)/k;
  n.non.positives <- length(non.positives);
  resto.non.positives <- n.non.positives%%k;
  n.non.pos.per.fold <- (n.non.positives - resto.non.positives)/k;
  j=1; 
  if (n.non.pos.per.fold > 0)
    for (i in 1:k) {
      fold.non.positives[[i]] <- non.positives[j:(j+n.non.pos.per.fold-1)];
      j <- j + n.non.pos.per.fold;
    }
  j.pos=1;  
  if (n.pos.per.fold > 0)
    for (i in 1:k) {
      fold.positives[[i]] <- positives[j.pos:(j.pos+n.pos.per.fold-1)];
      j.pos <- j.pos + n.pos.per.fold;
    }
  
  if (resto.non.positives > 0)
    for (i in k:(k-resto.non.positives+1)) {
      fold.non.positives[[i]] <- c(fold.non.positives[[i]], non.positives[j]);
      j <- j + 1;
    }
  
  if (resto.positives > 0) 
    for (i in 1:resto.positives) {
      fold.positives[[i]] <- c(fold.positives[[i]], positives[j.pos]);
      j.pos <- j.pos + 1;
    }
  
  return(list(fold.non.positives=fold.non.positives, fold.positives=fold.positives));
}




# Function to compute the F-measure for a single class
# Input
# pred : factor of the predicted labels
# labels : factor of the true labels
# Note that 0 level stands for negative and 1 for positive.
# In general the first level is negative and the second positive
# Output:
# The computed F-score
compute.F <- function(pred,labels) {      
     if (length(pred)!=length(labels))
         stop("compute.F: lengths of true and predicted labels do not match.");
     neg.ex <- which(labels <= 0);	
	 np <- length(neg.ex);
	 pos.ex <- which(labels > 0);
	 npos <- length(pos.ex);	 
	 TP <- sum(pred[pos.ex] > 0);
	 FN <- sum(pred[pos.ex] <= 0);	
	 TN <- sum(pred[neg.ex] <= 0);
	 FP <- sum(pred[neg.ex] > 0);	           
     if ((TP+FP) == 0)
       precision <- 0
     else 
       precision <- TP/(TP+FP);
     if ((TP+FN) == 0)
       recall <- 0
     else
       recall <- TP/(TP+FN);
     if ((precision+recall) == 0)
     	F <- 0
     else
     	F = 2 *(precision*recall) / (precision+recall);      
     return (F);
}

# Function to compute the accuracy for a single class
# Input
# pred : vector of the predicted labels
# labels : vector of the true labels
# Note that 0  stands for negative and 1 for positive.
# In general the first level is negative and the second positive
# Output:
# The computed accuracy
compute.acc <- function(pred,labels) {      
     if (length(pred)!=length(labels))
         stop("compute.acc: lengths of true and predicted labels do not match.");
     n <- length(labels);
	 acc <- sum(pred==labels)/n;
	 return(acc);
}
  

#.onLoad <- function(libname=.libPaths(), pkgname="RANKS")
#       library.dynam("ranks", pkgname, libname);
	   
#.onAttach <- function(libname=.libPaths(), pkgname="RANKS")
#              cat("RANKS: RAnking Nodes in Networks through Kernelized Score functions.\n");
